# vim: set fileencoding=utf-8 :

from django.db import models
from billing_models import AccountHistory, Operation, Bill, BillEntry, OperationType, OperationStatus
from datetime import datetime,date
from django.contrib.auth.models import User, Group
from django.conf import settings

from django.utils.translation import ugettext_lazy as _

from decimal import Decimal

the_app_label = 'katrin'

YES_NO = (
	(1, _('Yes')),
	(0, _('No')),
)

class Service(models.Model):
	id = models.AutoField(primary_key=True)
	title = models.CharField(_('Title'), max_length=120)
	codename =  models.CharField(_('Codename'), max_length=120, unique=True, help_text=_("You can't edit it. Katrin system uses it"))

	def __unicode__(self):
		return self.title
	class Meta:
		db_table = u'services'
		verbose_name = _('service')
		verbose_name_plural = _('services')
		app_label = the_app_label

# Abstract models
class Human(models.Model):
	id = models.AutoField(primary_key=True)
	login = models.CharField(_('Login'), max_length=60, unique=True)
	password = models.CharField(_('Password'), max_length=20)
	lastname = models.CharField(_('Lastname'), max_length=20)
	firstname = models.CharField(_('Firstname'), max_length=20)
	middlename = models.CharField(_('Middlename'), max_length=20)
	photo = models.FileField(blank=True, null=True, upload_to = "tmp")
	work_tel = models.CharField(_('Work tel'), max_length=50, null=True, blank=True)
	mobile_tel = models.CharField(_('Mobile phone'), max_length=50, null=True, blank=True)
	email = models.EmailField(_('E-mail'), max_length=50, null=True, blank=True)
	jabber = models.CharField(_('Jabber'), max_length=50, null=True, blank=True)
	user = models.ForeignKey(User, verbose_name=_('Django user'), unique=True, null=True, blank=True)

	class Meta:
		abstract = True
		app_label = the_app_label

	def __unicode__(self):
		return "%s %s %s (%s)" % (self.lastname, self.firstname, self.middlename, self.login)
	def save(self):
		if not self.user:
			user = User()
		else:
			user = self.user
		user.username = self.login
		user.set_password(self.password)
		user.first_name = self.firstname
		user.last_name = self.lastname
		if self.email:
			user.email = self.email
		else:
			user.email = u''
		if user.id == None:
			user.save()
		self.user = user

	def delete(self):
		if self.user:
			self.user.delete()

class Abonent(models.Model):
	"""Implements common things for Client and ClientGroup
	"""
	yu_name = models.CharField(_('Juridical name'), max_length=90, null=True, blank=True)
	yu_address = models.CharField(_('Juridical address'), max_length=150, null=True, blank=True)
	services = models.ManyToManyField(Service, through='UserService')

	def get_bill_to(self):
		if self.yu_name:
			return self.yu_name
		else:
			return u"%s" % self

	class Meta:
		abstract = True
		app_label = the_app_label

	@classmethod
	def is_valid(cls, abonent):
		if not abonent:
			raise ValueError("Abonent is None")
		elif not abonent.id:
			raise ValueError("Abonent hasn't been saved yet")
		else:
			return True

	def number_of_accounts(self):
		"""Returns number of accounts for given abonent or total number of accounts
		"""
		if not self:
			return Account.objects.all().count()
		return self.accounts().count()
	number_of_accounts.short_description = _("Number of accounts")

	def accounts(self):
		return Account.query_by_owner(self)

	def get_default_account(self):
		ac = self.accounts()[0]
		if not ac:
			raise ValueError("Can't find default account")
		else:
			return ac

	def has_accounts(self):
		if self.number_of_accounts() > 0:
			return True
		else:
			return False

	def save(self, admin=None):
		super(Abonent,self).save()
		if not self.has_accounts():
			# create default account
			ac = Account()
			ac.set_owner(self)
			ac.save(admin)

	def delete(self):
		for account in Account.owner_accounts(self):
			account.delete()
		super(Abonent,self).delete()

# end abstract models


class Tariff(models.Model):

	id = models.AutoField(primary_key=True)
	title = models.CharField(_('Title'), max_length=50)
	service = models.ForeignKey(Service, verbose_name=_('Service'))
	billing_period = models.ForeignKey('BillingPeriod', verbose_name=_('Billing period'), related_name="tariffs")
	activation_fee = models.DecimalField(_('Activation fee'), default=0, max_digits=10, decimal_places=4, help_text=_('Fee to start service usage'))
	periodic_fee = models.DecimalField(_('Periodic fee'), default=0, max_digits=10, decimal_places=4, help_text=_('Fee per billing period'))
	week_days = models.CharField(_('Days of week'), max_length=12, blank=True, help_text=_('Comma-separated number days of week (Example: 6,7). The tariff is active during these days.'))
	class Meta:
		db_table = u'tariffs'
		verbose_name = _('tariff')
		verbose_name_plural = _('Tariffs')
		app_label = the_app_label
	def __unicode__(self):
		return "%s %s %s %s" % (self.title, _('for'), _('service'), self.service)
	def has_this_service(self, service):
		if self.service == service:
			return True
		else:
			return False

class Client(Human,Abonent):
#	additional info

	class Meta:
		db_table = u'clients'
		verbose_name = _('client')
		verbose_name_plural = _('Clients')
		app_label = the_app_label

	def __unicode__(self):
		return u"%s" % (self.login)

	def member_of_group_with_service(self, service):
		return self.groups.filter(userservices__service=service).count() > 0

	def save(self, admin=None):
		Human.save(self)
		self.user.groups.add(Group.objects.get(name='clients'))
		self.user.save()
		Abonent.save(self,admin)
	
	def delete(self):
		Abonent.delete(self)
		Human.delete(self)

class ClientGroup(Abonent):
	""" Joins clients into groups

	Several clients can share the same account (accounts)
	"""
	clients = models.ManyToManyField(Client, verbose_name=_('Clients'), blank=True, null=True, related_name="groups")
	name = models.CharField(_('Name'), max_length=80)
	
	def __unicode__(self):
		return self.name

	def is_empty(self):
		return self.clients.count() == 0

	def members_ids(self):
		members = []
		for client in self.clients.all():
			members.append(client.pk)
		return members

	class Meta:
		db_table = u'clientgroups'
		verbose_name = _('client group')
		verbose_name_plural = _('Client groups')
		app_label = the_app_label

AccountType = {
	'Personal':	1,
	'Group':	2,
}

class Account(models.Model):
	"""Implements account model

	Each client and group of clients can have several accounts. Account has an owner who is either client or group 
	There are several types of accounts. Personal type is for client. Group type is for group.
	"""
	TYPE_CHOICES = (
			(AccountType['Personal'], _('Personal')),
			(AccountType['Group'], _('Group')),
	)

	name = models.CharField(_('Name'), max_length=30)
	blocked = models.IntegerField(_('Blocked'), default=0, choices=YES_NO)
	balance = models.DecimalField(_('Balance'), max_digits=10, decimal_places=4, default=Decimal('0'), help_text=_('Create operation to change balance'))
	credit = models.DecimalField(_('Credit'), max_digits=10, decimal_places=4, default=Decimal('0'))
	client = models.ForeignKey(Client, verbose_name=_('Client'), blank=True, null=True)
	group = models.ForeignKey(ClientGroup, verbose_name=_('Client group'), blank=True, null=True)
	type = models.IntegerField(_('Type'), default=AccountType['Personal'], choices=TYPE_CHOICES)
	postpaid = models.BooleanField(_('Postpaid'), default=False)
	ext_owner_id = models.CharField(_('ID in external system'), blank=True, null=True, max_length=30)

	class Meta:
		db_table = u'accounts'
		verbose_name = _('account')
		verbose_name_plural = _('Accounts')
		app_label = the_app_label

	# class methods
	@classmethod
	def query_by_owner(cls, owner):
		"""Returns queryset of accounts for given owner
		"""
		if isinstance(owner, Client):
			kwargs = {"client": owner,}
		elif isinstance(owner, ClientGroup):
			kwargs = {"group": owner,}
		else:
			raise ValueError("Invalid owner")
		return Account.objects.filter(**kwargs)

	# instance methods
	def get_owner(self):
		if self.type == AccountType['Personal']:
			return self.client
		elif self.type == AccountType['Group']:
			return self.group
		else:
			raise ValueError("No owner")

	get_owner.short_description = _('Owner')

	def set_owner(self, owner):
		if isinstance(owner, Client):
			self.type = AccountType['Personal']
			self.client = owner
			self.group = None
			self.name = owner
		elif isinstance(owner, ClientGroup):
			self.type = AccountType['Group']
			self.client = None
			self.group = owner
			self.name = owner
		else:
			raise ValueError("Wrong account owner")
	
	def close_billing_period(self, userservices, include_last_stat=False):
		from katrin.service import ServiceStat
		for us in userservices:
			stat_class = ServiceStat.get_stat_class(us.service.codename)
			#print stat_class
			if stat_class:
				stat_class.create_usage_operations(us, include_last_stat)
			Operation.do_periodic(us)
			Operation.do_closure(us)
			
		self.prepare_bill(userservices)

	def prepare_bill(self, userservices):
		bill_ops = Operation.objects.filter(account=self, service__in=map(lambda us: us.service, userservices), debit__gt=0, status=OperationStatus['No bill'])
		if len(bill_ops)>0:
			# We found operations and can prepare bill
			bill_to = self.get_owner().get_bill_to()
			bill = Bill(
					bill_to=bill_to,
					bill_to_extid=self.ext_owner_id
					)
			bill.save()

			for op in bill_ops:
				BillEntry(
						bill = bill,
						operation = op,
						net_amount = op.debit,
						vat_amount = op.debit*settings.VAT_PERCENT/100,
						).save()
				op.status=OperationStatus['Closed']
				op.save()


	def __unicode__(self):
		return self.name

	def save(self, admin=None, operation=None):
		if self.type == AccountType['Personal'] and Abonent.is_valid(self.client):
			self.group = None
		elif self.type == AccountType['Group'] and Abonent.is_valid(self.group):
			self.client = None
		else:
			raise ValueError("Unknown account type")
		ah = AccountHistory(
				date_time = datetime.now(),
				admin = admin,
				balance = self.balance,
				credit = self.credit,
				blocked = self.blocked,
				operation = operation,
			)
		super(Account, self).save()
		ah.account = self
		ah.save()


class UserService(models.Model):
	"""
	Joins User (Client/ClientGroup), Service, Tariff and Account

	Applies several restrictions:
	* unique services for given user
	* only one tariff plan for given service and user (Service,User->Tariff)
	* only one account for given service and user (Service,User->Account)
	* unique services for given account
	"""
	# useless because client and group can be NULL - unique doesn't work (SQL standart), see is_unique method
	# unique_together = ("client","group","service")

	client = models.ForeignKey(Client, verbose_name=_('Client'), blank=True, null=True, related_name='userservices')
	group = models.ForeignKey(ClientGroup, verbose_name=_('Client group'), blank=True, null=True, related_name='userservices')
	tariff = models.ForeignKey(Tariff, verbose_name=_('Tariff'), related_name='userservices')
	account = models.ForeignKey(Account, verbose_name=_('Account'), related_name='userservices')
	service = models.ForeignKey(Service, verbose_name=_('Service'), related_name='userservices') 

	class Meta:
		db_table = u'user_services'
		verbose_name = _('Service and tariff')
		verbose_name_plural = _('Services and tariffs')
		app_label = the_app_label

	def __unicode__(self):
		return u"%s: %s, %s: %s, %s: %s" % (_('Service'), self.service, _('Tariff'), self.tariff, _('Account'), self.account, )

	def user(self):
		if self.client and Abonent.is_valid(self.client):
			return self.client
		elif self.group and Abonent.is_valid(self.group):
			return self.group
		else:
			raise ValueError("Can't find user of service")
	
	def is_unique(self):
		if UserService.objects.filter(client=self.client, group=self.group, service=self.service).count() > 0:
			return False
		else:
			return True

	def is_valid(self):
		if self.client:
			# We don't need clients inside group with the same service
			if self.client.member_of_group_with_service(self.service):
				return False
		elif self.group:
			# We don't need empty groups
			if self.group.is_empty():
				return False
		return True

	def get_previous_period_closure(self):
		closure =  Operation.objects.filter(account=self.account, type=OperationType['Closure'], service=self.service).order_by('-date_time')[:1]
		if closure:
			return closure[0]
		else:
			return None
	
	def save(self, admin=None):
		if self.client: 
			self.group = None
		elif self.group:
			self.client = None
		user = self.user()
		try:
			self.account
		except(models.ObjectDoesNotExist):
			self.account = user.get_default_account()
		finally:
			if self.tariff.has_this_service(self.service):
				if self.is_unique():
					# activate new service and charge activation fee
					if not self.pk:
						Operation.do_activation(self)
					super(UserService,self).save()
				else:
					# we edit existing service
					if self.pk:
						super(UserService,self).save()
					else:
						raise ValueError("Service for given user already exists")
			else:
				raise ValueError("Given tariff isn't for given service")

